// Snakes for TI-89
// (C) 2003-2005 Jeremy Stanley <jstanley0@gmail.com>

// This program is licensed under the GNU General Public License
// see COPYING for details

#define USE_TI89              // Compile for TI-89
#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
#define MIN_AMS 100           // Compile for AMS 1.00 or higher
#define SAVE_SCREEN           // Save/Restore LCD Contents
#define NO_EXIT_SUPPORT				// To shave off a couple of bytes

#include <tigcclib.h>         
#include "util.h"

// FPS counter
// if defined, we'll count FPSs
// note that the timer relies on auto int 5, so keyboard handling with _rowread
// might not work correctly. **don't enable this** in release builds
// (_rowread sometimes gets false positives on ESC if the interrupts are enabled)
//#define FRAMECOUNTER

// game constants

#define P_LIVES 5
#define P_MAX_LIVES 8
#define P_IV_TIME 50
#define P_INIT_GOAL 50
#define P_GOAL_INC  50
#define P_MAX_GOAL 32750
#define SKILL_DELAY_FACTOR 6000

// world dimensions
#define W_WIDTH   160
#define W_HEIGHT   82
#define W_TOP_BORDER 8
#define W_BOTTOM_BORDER 10
#define W_DOOR_LEFT 70
#define W_DOOR_WIDTH 20

// player dimensions
#define P_WIDTH  8
#define P_HEIGHT 8
#define P_MIN_X 1
#define P_MIN_Y W_TOP_BORDER
#define P_MIN_Y_OPEN 0
#define P_MAX_X (W_WIDTH - P_WIDTH - 1)
#define P_MAX_Y (W_TOP_BORDER + W_HEIGHT - P_HEIGHT)
#define P_MAX_Y_OPEN (W_TOP_BORDER + W_HEIGHT + W_BOTTOM_BORDER - P_HEIGHT)

// treasure dimensions
#define T_WIDTH 8
#define T_HEIGHT 8
#define T_ORIG_COUNT 2
#define T_MAX_COUNT 16
#define T_MIN_X 1
#define T_MIN_Y W_TOP_BORDER
#define T_MAX_X (W_WIDTH - T_WIDTH - 1)
#define T_MAX_Y (W_TOP_BORDER + W_HEIGHT - T_HEIGHT)

// snake block dimensions
#define S_WIDTH 2
#define S_HEIGHT 2
#define S_ORIG_SNAKES 1
#define S_MAX_SNAKES 16
#define S_MIN_SNAKE_LEN 24 
#define S_SNAKE_LEN_INC  4
#define S_MAX_SNAKE_LEN 64
#define S_MIN_X 1
#define S_MIN_Y W_TOP_BORDER
#define S_MAX_X (W_WIDTH - S_WIDTH - 1)
#define S_MAX_Y (W_TOP_BORDER + W_HEIGHT - S_HEIGHT)

// player sprite
static unsigned char player_sprite[P_HEIGHT] =
{
	0b00111100,
	0b00111100,
	0b00011000,
	0b11111111,
	0b00011000,
	0b00111100,
	0b01100110,
	0b11000011
};

// treasure sprite
static unsigned char treasure_sprite[T_HEIGHT] = 
{
	0b00000000,
	0b01111110,
	0b10100001,
	0b11111111,
	0b10101101,
	0b10100001,
	0b11111111,
	0b00000000
};

// Main Function
void _main(void)
{
	// globals
	int skill;
	int delay;
	int dc;
	int goal;
	int endgame;
	int level;
	int lives;
	int iv_time;
	int score;
	int top_open;
	int bottom_open;
	LCD_BUFFER buf;
	int old_font;
		
	short player_coords[2];
	
	int snake_count;
	short snake_coords[S_MAX_SNAKES][2];
	short snake_dirs[S_MAX_SNAKES][2];
	
	int treas_count;
	int vis_treas_count;
	short treas_coords[T_MAX_COUNT][2];
	short treas_exists[T_MAX_COUNT];
	
	INT_HANDLER oldHandler1;
#ifdef FRAMECOUNTER
	int frames = 0;
#else
	INT_HANDLER oldHandler5;
#endif
	
	short show_title_screen()
	{
		short key;
	
		ClrScr();
		
		center_string(16, 0, "Snakes", F_8x10, A_NORMAL);
		center_string(28, 0, "v1.2 (built 1/18/2004)", F_4x6, A_NORMAL);
		center_string(50, 0, "by Jeremy Stanley", F_6x8, A_NORMAL);
		center_string(60, 0, "jstanley0@gmail.com", F_4x6, A_NORMAL);
		center_string(82, 0, "Select skill level to begin", F_4x6, A_NORMAL);
		center_string(90, 0, "F1=Easy ... F5=Impossible", F_4x6, A_NORMAL);

		while(1) 
		{
			switch(key=ngetchx())
			{
			  case KEY_F1 ... KEY_F5:
			  	return (key - KEY_F1) + 1;
			  case KEY_ESC:
			  	return 0;
		//	  case 'x':
		//  		return -1;
			}
		}
	}
	
	void update_level_str()
	{
		char level_str[32];
		
		sprintf(level_str, "level=%d", level);
		
		DrawStr(1, 1, level_str, A_REPLACE);
	}
	
	void update_score_str()
	{
		char score_str[32];
		int score_x;		
		
		sprintf(score_str, "score=%d", score);
		score_x = W_WIDTH - DrawStrWidth(score_str, F_4x6) - 1;
		
		DrawStr(score_x, 1, score_str, A_REPLACE);			
	}
	
	void init_snakes()
	{
		int i;
		for(i = 0; i < snake_count; i++)
		{
			snake_coords[i][0] = random(S_MAX_X - S_MIN_X) + S_MIN_X;
			snake_coords[i][1] = random(S_MAX_Y - S_MIN_Y) + S_MIN_Y;

			snake_dirs[i][0] = (rand() & 1) ? -1 : 1;
			snake_dirs[i][1] = (rand() & 1) ? -1 : 1;
		}
	}
	
	void init_treasures()
	{
		int i;
		for(i = 0; i < treas_count; i++)
		{
			treas_coords[i][0] = random(T_MAX_X - T_MIN_X) + T_MIN_X;
			treas_coords[i][1] = random(T_MAX_Y - T_MIN_Y) + T_MIN_Y;
			treas_exists[i] = 1;
		}
	}
	
	void update_snakes()
	{
		int i;
		for(i = 0; i < snake_count; i++)
		{
			if (snake_coords[i][0] == S_MAX_X)
				snake_dirs[i][0] = -1;
			else if (snake_coords[i][0] == S_MIN_X)
				snake_dirs[i][0] = 1;
					
			if (snake_coords[i][1] == S_MAX_Y)
				snake_dirs[i][1] = -1;
			else if (snake_coords[i][1] == S_MIN_Y)
				snake_dirs[i][1] = 1;

			snake_coords[i][0] += snake_dirs[i][0];
			snake_coords[i][1] += snake_dirs[i][1];
		}
	}
	
	void next_level()
	{
		if (++level > 1)
		{
			if (++treas_count > T_MAX_COUNT)
				treas_count = T_MAX_COUNT;
			
			if (++snake_count > S_MAX_SNAKES)
				snake_count = S_MAX_SNAKES;
		}
		vis_treas_count = treas_count;
		iv_time = 0;
		top_open = 0;
		bottom_open = 1;
		player_coords[1] = W_TOP_BORDER + W_HEIGHT + W_BOTTOM_BORDER/2 - P_HEIGHT/2;
		init_snakes();
		init_treasures();
		
		// clear and initialize score area
		erase_rect(0, 0, W_WIDTH, W_TOP_BORDER, buf);
		update_level_str();
		update_score_str();
	}

	void init_game()
	{
		player_coords[0] = W_DOOR_LEFT + W_DOOR_WIDTH/2 - P_WIDTH/2;
		delay = (5 - skill) * SKILL_DELAY_FACTOR;
		goal = P_INIT_GOAL;
		score = 0;
		endgame = 0;
		level = 0;
		lives = P_LIVES;
		treas_count = T_ORIG_COUNT;
		snake_count = S_ORIG_SNAKES;
		randomize();
		next_level();
	}
	
	void draw_frame()
	{
#ifdef FRAMECOUNTER
		static char fbuf[8] = "---";
		frames++;
		if(OSTimerExpired(USER_TIMER))
		{
			sprintf(fbuf, "%03d", frames);
			frames = 0;
		}
		DrawStr(W_WIDTH - DrawStrWidth(fbuf, F_4x6) - 1, 94, fbuf, A_REPLACE);
#endif	

		// top
		if (top_open)
		{
			hline(0, W_TOP_BORDER - 1, W_DOOR_LEFT, buf);
			vline(W_DOOR_LEFT - 1, 0, W_TOP_BORDER, buf);
			vline(W_DOOR_LEFT + W_DOOR_WIDTH, 0, W_TOP_BORDER, buf);
			hline(W_DOOR_LEFT + W_DOOR_WIDTH, W_TOP_BORDER - 1, W_WIDTH - W_DOOR_WIDTH - W_DOOR_LEFT, buf);
		}
		else
		{
			hline(0, W_TOP_BORDER - 1, W_WIDTH, buf);
		}
		
		// sides
		vline(0, W_TOP_BORDER, W_HEIGHT, buf);
		vline(W_WIDTH - 1, W_TOP_BORDER, W_HEIGHT, buf);
		
		// bottom
		if (bottom_open)
		{
			hline(0, W_HEIGHT + W_TOP_BORDER, W_DOOR_LEFT, buf);
			vline(W_DOOR_LEFT - 1, W_HEIGHT + W_TOP_BORDER, W_BOTTOM_BORDER, buf); 
			vline(W_DOOR_LEFT + W_DOOR_WIDTH, W_HEIGHT + W_TOP_BORDER, W_BOTTOM_BORDER - 1, buf);
			hline(W_DOOR_LEFT + W_DOOR_WIDTH, W_HEIGHT + W_TOP_BORDER, W_WIDTH - W_DOOR_WIDTH - W_DOOR_LEFT, buf);
		}
		else
		{
			hline(0, W_HEIGHT + W_TOP_BORDER, W_WIDTH, buf);
		}
	}

	inline void draw_snake_segment(int x, int y)
	{
		unsigned char *p = pixel_ptr(x, y, buf);
		unsigned char mask = 0xC0 >> (x & 7);
		
		*p |= mask;
		*(p + 30) |= mask;
		
		if (mask == 1)
		{
			*(p + 1) |= 0x80;
			*(p + 31) |= 0x80;
		}
	}
	
	void draw_snakes()
	{
		int i, l;
		int x, y, dx, dy;
		int snake_len = S_MIN_SNAKE_LEN;
		
		for(i = 0; i < snake_count; i++)
		{
			x = snake_coords[i][0];
			y = snake_coords[i][1];
			dx = -snake_dirs[i][0];
			dy = -snake_dirs[i][1];
			
			for(l = 0; l < snake_len; l++)
			{
				if (x == S_MAX_X)
					dx = -1;
				else if (x == S_MIN_X)
					dx = 1;
					
				if (y == S_MAX_Y)
					dy = -1;
				else if (y == S_MIN_Y)
					dy = 1;

				draw_snake_segment(x, y);
																
				x += dx;
				y += dy;
			}
		  snake_len += S_SNAKE_LEN_INC;
		  if ( snake_len > S_MAX_SNAKE_LEN )
			  snake_len = S_MAX_SNAKE_LEN;
		}
	}

	// with apologies to David McLaughlin
	void draw_player_and_test_player_collission_with_snakes()
	{
		if (iv_time == 0)
		{
			if (sprite8_draw_ctest(player_coords[0], player_coords[1], P_HEIGHT, player_sprite, buf))
			{
				if (--lives == 0)
				{
					endgame = 1;
				}	
				else
				{
					iv_time = P_IV_TIME;				
				}
			}
		}
		else
		{
			// collission test not necessary when invulnerable
			// also, eh, let's make the dude blink
			if (iv_time & 4)
				sprite8(player_coords[0], player_coords[1], P_HEIGHT, player_sprite, buf);
		}
	}
	
	void draw_treasures_and_test_player_collissions_with_them()
	{
		int i;
		for(i = 0; i < treas_count; i++)
		{
			if (treas_exists[i])
			{
				if (rects_overlap(player_coords, P_WIDTH, P_HEIGHT,
				                  treas_coords[i], T_WIDTH, T_HEIGHT))
				{
					score += skill;
					if (score > goal)
					{
						if (goal < P_MAX_GOAL) goal += P_GOAL_INC;
						if (lives < P_MAX_LIVES) lives++;						
					}
					treas_exists[i] = 0;
					if (--vis_treas_count == 0)
						top_open = 1;
						
					update_score_str();
				}
				else
				{
					sprite8(treas_coords[i][0], treas_coords[i][1], T_HEIGHT, treasure_sprite, buf);
				}
			}
		}		
	}

	void draw_reserve_players()
	{
		int i;
		int x = 0, y = W_TOP_BORDER + W_HEIGHT + 2;
		
		for(i = 1; i < lives; i++)
		{
			sprite8(x, y, P_HEIGHT, player_sprite, buf);
			x += P_WIDTH + 2;
		}
	}

	void draw_game()
	{
		// don't clear the top rows, including score and level.
		// it's expensive to draw that text, so we only do it when we need to.
		// however, make sure we _do_ clear the top of the box
		memset(buf + (W_TOP_BORDER * 30), 0, ((W_HEIGHT + W_BOTTOM_BORDER) * 30));
		erase_rect(W_DOOR_LEFT, 0, W_DOOR_WIDTH, W_TOP_BORDER, buf);
			
		draw_snakes();
		draw_player_and_test_player_collission_with_snakes();
		draw_reserve_players();
		draw_frame();
		draw_treasures_and_test_player_collissions_with_them();
	}

	void move_player(short dx, short dy)
	{
		short minx = P_MIN_X;
		short maxx = P_MAX_X;
		short miny = P_MIN_Y;
		short maxy = P_MAX_Y;

		// close entry door once clear
		if (player_coords[1] < W_TOP_BORDER + W_HEIGHT - P_HEIGHT)
			bottom_open = 0;

		// horizontal movement
		
		if ((top_open && (player_coords[1] < W_TOP_BORDER)) ||
		    (bottom_open && (player_coords[1] + P_HEIGHT > W_TOP_BORDER + W_HEIGHT)))
	  {
	  	// inside one of the doors
	  	minx = W_DOOR_LEFT; 
	  	maxx = W_DOOR_LEFT + W_DOOR_WIDTH - P_WIDTH;
	  }

		player_coords[0] += dx;
		if (player_coords[0] < minx)
			player_coords[0] = minx;
		else if (player_coords[0] > maxx)
			player_coords[0] = maxx;

		// vertical movement
		
		if ((player_coords[0] >= W_DOOR_LEFT) && (player_coords[0] <= W_DOOR_LEFT + W_DOOR_WIDTH - P_WIDTH))
		{
			if (top_open)
				miny = P_MIN_Y_OPEN;
			if (bottom_open)
				maxy = P_MAX_Y_OPEN;	
		}
						
		player_coords[1] += dy;
		if (player_coords[1] < miny)
			player_coords[1] = miny;
		else if (player_coords[1] > maxy)
			player_coords[1] = maxy;
	}
	
	void scroll_to_next_level()
	{
		int h0, h;
		unsigned char *pold;
	
		draw_game();
		
		// erase score, level
		erase_rect(0, 0, W_DOOR_LEFT - 2, W_TOP_BORDER - 1, LCD_MEM);
		erase_rect(W_DOOR_LEFT + W_DOOR_WIDTH + 1, 0, W_WIDTH - W_DOOR_LEFT - W_DOOR_WIDTH, W_TOP_BORDER - 1, LCD_MEM);

		// do a spiffy scrolly thingy from LCD memory to buf	
		h0 = W_TOP_BORDER + W_HEIGHT;
		pold = buf + h0 * 30;
		
		for(h = h0; h >= 0; h--)
		{
			int i;
			for(i = 99; i >= 0; i--)
			{
				memcpy(LCD_MEM + ((i + 1) * 30), LCD_MEM + (i * 30), 30);
			}
			
			memcpy(LCD_MEM, pold, 30);
			pold -= 30;
		}
	}
	
	// main
	skill = show_title_screen();
	if (!skill)
		return;

	// set virtual display buffer
	PortSet(buf, 239, 127);
	old_font = FontSetSys(F_4x6);
	
	// clear auto-ints 1 and 5 so that _keytest works properly
	oldHandler1 = GetIntVec(AUTO_INT_1);
	SetIntVec(AUTO_INT_1, DUMMY_HANDLER);
#ifdef FRAMECOUNTER
	OSRegisterTimer(USER_TIMER, 20);
#else
	oldHandler5 = GetIntVec(AUTO_INT_5);
	SetIntVec(AUTO_INT_5, DUMMY_HANDLER);
#endif

	init_game();
	endgame = 0;
		
	do
	{
		short dx = 0, dy = 0;
		
		BEGIN_KEYTEST
			if (_keytest_optimized(RR_LEFT))
				dx = -1;
			else if (_keytest_optimized(RR_RIGHT))
				dx = 1;
			if (_keytest_optimized(RR_UP))
				dy = -1;
			else if (_keytest_optimized(RR_DOWN))
				dy = 1;
		END_KEYTEST
				
		if (_keytest(RR_ESC)) 
		{
			endgame = 1;
		}
		
		move_player(dx, dy);

		update_snakes();
		
		draw_game();
		LCD_restore(buf);
		
		if (iv_time > 0)
		{
			--iv_time;
		}
	
		for(dc = 0; dc < delay; dc++);
//idle();
	
		if (player_coords[1] == P_MIN_Y_OPEN)
		{
			next_level();
			scroll_to_next_level();
		}
	}
	while(!endgame);
	
	// restore interrupt handlers
	SetIntVec(AUTO_INT_1, oldHandler1);
#ifdef FRAMECOUNTER
  OSFreeTimer(USER_TIMER);
#else
	SetIntVec(AUTO_INT_5, oldHandler5);
#endif
	
	FontSetSys(F_6x8);
	DrawStr(0, W_TOP_BORDER + W_BOTTOM_BORDER + W_HEIGHT - 8, "GAME OVER", A_REPLACE);
	invert_lcd(buf);
	LCD_restore(buf);

	while(KEY_ESC != ngetchx());
	
	// restore original display buffer
	FontSetSys(old_font);
	PortRestore();
}
