// util.cpp -- graphic and sprite related utilities

// (C) 2003-2005 Jeremy Stanley <jstanley0@gmail.com>
// This program is licensed under the GNU General Public License
// see COPYING for details

#include <graph.h> 
#include <compat.h>
#include "util.h"

// eh?  if I use LCD_WIDTH from <compat.h> , I get a linker error
#define EH_LCD_WIDTH 160

int rects_overlap(short r1x[2], short w1, short h1, short r2x[2], short w2, short h2)
{
	if (r1x[0] + w1 <= r2x[0])
		return 0;	// r1 left of r2
	if (r1x[1] + h1 <= r2x[1])
		return 0;	// r1 above r2
	if (r2x[0] + w2 <= r1x[0])
		return 0; // r2 left of r1
	if (r2x[1] + h2 <= r1x[1])
		return 0; // r2 above r1
		
	return 1;
}
				                  
void invert_lcd(void *buf)
{
	unsigned int *lcd = (unsigned int *)buf;
	unsigned int *end = lcd + (LCD_SIZE / sizeof(unsigned int));
	
	while(lcd != end)
	{
		*lcd ^= ~0;
		lcd++;
	}
}

void sprite8(short x, short y, short h, unsigned char *sp, void *scr)
{
	int i;
	unsigned char *top = pixel_ptr(x, y, scr);
	unsigned char *p = top;
	short rem = (x & 7);

	for(i = 0; i < h; i++)
	{
		*p |= (sp[i] >> rem);
		p += 30;
	}

	if (rem)
	{
		p = top + 1;
		rem = 8 - rem;
		for(i = 0; i < h; i++)
		{
			*p |= (sp[i] << rem);		
			p += 30;
		}		
	}
}

int sprite8_draw_ctest(short x, short y, short h, unsigned char *sp, void *scr)
{
	int i;
	int coll = 0;
	unsigned char *top = pixel_ptr(x, y, scr);
	unsigned char *p = top;
	short rem = (x & 7);

	for(i = 0; i < h; i++)
	{
		unsigned char mask = sp[i] >> rem;
		if (*p & mask)
			coll = 1;
		*p |= mask;
		p += 30;
	}

	if (rem)
	{
		p = top + 1;
		rem = 8 - rem;
		for(i = 0; i < h; i++)
		{
			unsigned char mask = sp[i] << rem;
			if (*p & mask)
				coll = 1;		
			*p |= mask;		
			p += 30;
		}		
	}

	return coll;
}

void center_string(short row, short dx, const char *str, short font, short attr)
{
	short x = ((EH_LCD_WIDTH - DrawStrWidth(str, font)) >> 1) + dx;
	FontSetSys(font);
	DrawStr(x, row, str, attr);
}


// fast vertical line function
void vline(short x, short y, short h, void *buf)
{
	unsigned char *p = pixel_ptr(x, y, buf);
	unsigned char mask = pixel_mask(x);
	while(h > 0)
	{
		*p |= mask;
		p += 30;
		h--;
	}
}

// fast horizontal line function (writes up to 16 pixels at a time! w00t!)
void hline(short x, short y, short w, void *buf)
{
	unsigned short *p_row = (unsigned short *)buf + (y * 15);
	unsigned short *p_beg = p_row + (x >> 4);
	unsigned short *p_end = p_row + ((x + w) >> 4);
	unsigned short l_mask = 0xFFFFU >> (x & 0xF);
	unsigned short r_mask = ~(0xFFFFU >> ((x + w) & 0xF));
	if (p_beg == p_end)
	{
		*p_beg |= (l_mask & r_mask);
	}
	else
	{
		*p_beg++ |= l_mask;
		while(p_beg < p_end)
		{
			*p_beg++ = 0xFFFFU;
		}
		*p_end |= r_mask;
	}
}

// draw a rectangle...
void draw_rect(short x, short y, short w, short h, void *buf)
{
	unsigned short *p_row = (unsigned short *)buf + (y * 15);
	unsigned short l_mask = 0xFFFFU >> (x & 0xF);
	unsigned short r_mask = ~(0xFFFFU >> ((x + w) & 0xF));	
	unsigned short l_rem = (x >> 4);
	unsigned short r_rem = ((x + w) >> 4);

	while(h > 0)
	{
		unsigned short *p_beg = p_row + l_rem;
		unsigned short *p_end = p_row + r_rem;

		if (p_beg == p_end)
		{
			*p_beg |= (l_mask & r_mask);
		}
		else
		{
			*p_beg++ |= l_mask;
			while(p_beg < p_end)
			{
				*p_beg++ = 0xFFFFU;
			}
			*p_end |= r_mask;
		}
		
		p_row += 15;		
		h--;
	}
}

// erase a rectangle...
void erase_rect(short x, short y, short w, short h, void *buf)
{
	unsigned short *p_row = (unsigned short *)buf + (y * 15);
	unsigned short l_mask = ~(0xFFFFU >> (x & 0xF));
	unsigned short r_mask = 0xFFFFU >> ((x + w) & 0xF);	
	unsigned short l_rem = (x >> 4);
	unsigned short r_rem = ((x + w) >> 4);

	while(h > 0)
	{
		unsigned short *p_beg = p_row + l_rem;
		unsigned short *p_end = p_row + r_rem;

		if (p_beg == p_end)
		{
			*p_beg &= (l_mask | r_mask);
		}
		else
		{
			*p_beg++ &= l_mask;
			while(p_beg < p_end)
			{
				*p_beg++ = 0;
			}
			*p_end &= r_mask;
		}
		
		p_row += 15;		
		h--;
	}
}